//=============================================================================
// (c) Copyright 2009 Diamond Systems Corporation. Use of this source code
// is subject to the terms of Diamond Systems' Software License Agreement.
// Diamond Systems provides no warranty of proper performance if this
// source code is modified.
//
// File: eepconf.c   v2.1
// Using: DSCUD v6.00 
// Description: configures EMM-8, EMM-8P, and Neptune SBC serial ports.
// Created by KL, James Moore
//=============================================================================

#include <stdio.h>
#include <stdlib.h>


// Neptune register definitions
#define MISC_CONTROL_REGISTER           8 // Controls the PAGE, Reset and Interrupt functions
#define EEPROM_ACCESS_REGISTER         15 // Write 0xA5 to this location to enable enhanced features and access to the EEPROM; 0xA6 disables.
#define EEPROM_ADDRESS_REGISTER        20 // Address that the EEPROM pointer points to.
#define EEPROM_DATA_REGISTER           21 // Data that the EEPROM will write.
#define CALIBRATION_CONTROL_REGISTER   14 // Has bits to signal FPGA to write to EEPROM.

// Basically, all of these registers are the EMM-8P registers plus 16
#define EEPROM_BUSY_STATUS             20 // Bit 7 shows whether the EEPROM is busy or not.

#define POINTER_ENABLE_REGISTER        16 // Enables the ports after pointer is used to program modes into following register.
#define ADDRESS_IRQ_PROTOCOL_REGISTER  17 // Data to be written to volatile configuration memory.

#ifdef __BORLANDC__
#include <conio.h>
#include <dos.h>
#endif

#if defined(linux)
#include <sys/io.h>
#include <unistd.h>

void delay (int value)
{
    usleep(value * 1000 );
}

void outp(unsigned short addr, int value)
{
    outb(value, addr);
}

int inp(unsigned short addr)
{
    return (int)(inb(addr));
}

#endif


#ifdef _WIN32
#include "dscud.h"

void delay(int value)
{
    dscSleep(value);
}

void outp(unsigned short addr, int value)
{
    dscOutp(addr, value);
}

int inp(unsigned short addr)
{
    BYTE value;
    dscInp(addr, &value);
    return (int)value;
}

#endif


#define MAXPORTS 8
#define TIMEOUT 2000

int ports_variable = 8;

#define SERIAL_RS232     0
#define SERIAL_RS422     1
#define SERIAL_RS485     3

//a port is defined by its address and interrupt
struct port {
    int address;
    int intr;
    int serialproto;
};

int commandline = 0; //set to 1 if using commandline options

int configured = 0; //Set to 1 once the data structure below is configured

//array of 8 ports corresponding to those on the board.
//stores the address and interrupt of port x in ports[x].
port ports[MAXPORTS] = { {0,0,0} ,
                         {0,0,0} ,
                         {0,0,0} ,
                         {0,0,0} ,
                         {0,0,0} ,
                         {0,0,0} ,
                         {0,0,0} ,
                         {0,0,0} };

int base_address = 0; //base address of the board
int board_type = 0; // menu selection: EMM8 = 0, EMM8P = 1, Neptune = 2

// All comments which mention base_address+HARD_CODED_VALUE are referring to original base addresses of inherited
// serial ports register locations in the EMM-8 and EMM-8P. The #define constants above supersede these
// (if curious, look up original register definition in the respective manual).

int writeEEPROM(int addr, int data)
{
    int timeout = TIMEOUT;


	if ( board_type == 2 )
	{

		outp ( base_address + EEPROM_DATA_REGISTER, data );
		outp ( base_address + EEPROM_ADDRESS_REGISTER, addr | 0x80 );// to indicate Write operation

		// 32 corresponds to bit 7, the EEPROM is busy bit.
		while ( inp( base_address + EEPROM_BUSY_STATUS ) >= 128 && timeout) {delay(1); timeout--;}
		if (timeout == 0) return 0;
	}
	else
	{
		outp(base_address+5,data);
		outp(base_address+4,addr+128);

		while (inp(base_address+4) >= 128 && timeout) {delay(1); timeout--;}
	}


    #ifdef __BORLANDC__
    delay(100);
    #endif

    return 1;
}

/*
reads in the value at the eeprom's "addr" and stores it to the
integer pointed to by "val".  assumes the global "base_address" is set
to the eem8's base adddress.

1.  write addr to base_address+4
2.  wait for bit 7 of base_address+4 to go to 0
3.  read base_address+5 into *val
4.  done

returns 1 on success, 0 on a timeout
*/

int readEEPROM(int addr, int * val)
{
    int timeout = TIMEOUT;

	if ( board_type == 2 )
	{
		outp ( base_address + EEPROM_ADDRESS_REGISTER, addr );

		// 32 corresponds to bit 7, the EEPROM is busy bit.
		while ( inp ( base_address + EEPROM_BUSY_STATUS ) >= 128 && timeout) {delay(100); timeout-= 100;}
	}
    else
    {
		outp(base_address+4, addr);
		while (inp(base_address+4) >= 128 && timeout) {delay(100); timeout-= 100;}
	}
    #ifdef __BORLANDC__
    delay(100);
    #endif

    if (timeout <= 0) return 0;


	if ( board_type == 2 )
	{
	    *val = (int)inp ( base_address + EEPROM_DATA_REGISTER );
	}
	else
	{
    	*val = (int)inp(base_address+5);
	}

    return 1;
}


int reconfigEMM8FromMemory()
{
    int proto1, proto2;
    int x = 0;
	int retval = 0;

    proto1 = 0;
    proto2 = 0;

    printf("Reconfiguring board...");


	if ( board_type != 0 )
	{

    //=======NEW for EMM8P======//
    for (x = 0; x < ports_variable; x++)
    {
        if ( x < 4 )
            proto1 |= (ports[x].serialproto) << (x * 2);
        else
            proto2 |= (ports[x].serialproto) << ((x-4) * 2);
    }

    //load serial protocols
    //this will only work for EMM8P and should not work for regular EMM8
	if ( board_type == 2 )
	{
		outp ( base_address + POINTER_ENABLE_REGISTER, 8 | 0x80 );
		outp ( base_address + ADDRESS_IRQ_PROTOCOL_REGISTER, proto1 );
	}
	else
	{
		outp(base_address, 16 | 0x80);
		outp(base_address+1, proto1);

		outp(base_address, 17 | 0x80);
		outp(base_address+1, proto2);
	}

    //===END new for EMM8P===//
	}


    for (x = 0; x < ports_variable; x++)
    {
        if (ports[x].address != 0)
        {
		if ( board_type == 2 )
		{
            //output port # to base_address
            //then output address >> 3 to base_address+1
            outp ( base_address + POINTER_ENABLE_REGISTER, x );
            outp ( base_address + ADDRESS_IRQ_PROTOCOL_REGISTER, ports[x].address >> 3 );

            //output port # with bit 3 set to 1 to base_address
            //then output the interrupt to base_address+1
            outp ( base_address + POINTER_ENABLE_REGISTER, x+4 );
            outp ( base_address + ADDRESS_IRQ_PROTOCOL_REGISTER, ports[x].intr );
        }
        else
        {
			//output port # to base_address
			//then output address >> 3 to base_address+1
			outp(base_address, x);
			outp(base_address+1, ports[x].address >> 3);

			//output port # with bit 3 set to 1 to base_address
			//then output the interrupt to base_address+1
			outp(base_address, x+8);
			outp(base_address+1, ports[x].intr);
		}
        }
    }
    //after all ports configured, set enable bit on base_address (bit 7) to 1

	if ( board_type == 2 )
    {
		outp(base_address + POINTER_ENABLE_REGISTER,0x80);
	    retval = inp ( base_address + ADDRESS_IRQ_PROTOCOL_REGISTER );
	}
	else
	{
		outp(base_address,0x80);
		retval = inp(base_address+1);
	}

    //to see if configured correctly, read from base_address+1
    //and it should be the address of port zero >> 3
    if ( (retval << 3) != ports[0].address)
    {
        printf("FAILED\n\n");
        return 0;
    }


    printf("Done\n\n");
    return 1;
}

void saveConfigToFile(char* filename)
{
    FILE* file = fopen(filename, "wb");
    if (file == NULL) { printf("Could not save file\n"); fclose(file); return; }

    fprintf(file, "BASEADDRESS=%x\r\n", base_address);

    for (int p = 0; p < ports_variable; p++)
    {
        fprintf(file,"[PORT %d]\r\n",p+1);
        fprintf(file,"ADDRESS=%x\r\n",ports[p].address);
        fprintf(file,"INTERRUPT=%d\r\n",ports[p].intr);
        fprintf(file,"SERIALPROTOCOL=%d %s\r\n", ports[p].serialproto, (ports[p].serialproto == SERIAL_RS232)?"RS232": ((ports[p].serialproto == SERIAL_RS422)?"RS422": "RS485") );
    }

    fclose(file);

    printf("Configuration Saved\n\n");
}

void loadConfigFromFile(char * filename)
{
    int buf;
    char temp[256];

    FILE* file = fopen(filename, "rb");
    if (file == NULL) { printf("Could not open file\n"); fclose(file); return;}

    fscanf(file, "BASEADDRESS=%x\r\n", &base_address);



    for (int p = 0; p < ports_variable; p++)
    {
        fscanf(file, "[PORT %d]\r\n", &buf);
        fscanf(file, "ADDRESS=%x\r\n", &(ports[p].address));
        fscanf(file, "INTERRUPT=%d\r\n", &(ports[p].intr));
        fscanf(file, "SERIALPROTOCOL=%d %s\r\n", &(ports[p].serialproto), temp);
    }

    fclose(file);

     printf("Configuration loaded to memory\n\n");
    //if (!commandline) reconfigEMM8FromMemory();

    configured = 1;
}

//writeConfigToEEPROM fills 16 bytes of EEPROM
//each pair of bytes stores the address >> 3 and the interrupt
//of a port.
void writeConfigToEEPROM()
{
    int proto1, proto2, x;

    proto1 = 0;
    proto2 = 0;

    if (!configured)
    {
        printf("No configuration in memory!\n");
      return;
    }

    printf("Writing configuration to EEPROM");

	if ( board_type != 0 )
	{
		//=======NEW for EMM8P======//
		for (x = 0; x < ports_variable; x++)
		{
			if ( x < 4 )
				proto1 |= (ports[x].serialproto) << (x * 2);
			else
				proto2 |= (ports[x].serialproto) << ((x-4) * 2);
		}

		if ( board_type == 2 )
		{
			if (!writeEEPROM(40,proto1)) { printf("\nEEPROM write error:  write operation timed out.\n\n"); return; }
		}
		else
		{
			if (!writeEEPROM(16,proto1)) { printf("\nEEPROM write error:  write operation timed out.\n\n"); return; }
			if (!writeEEPROM(17,proto2)) { printf("\nEEPROM write error:  write operation timed out.\n\n"); return; }
		}

		//=======END NEW for EMM8P======//
	}

    for (x = 0; x < ports_variable; x++)
    {
		if ( board_type == 2 )
		{
			if (!writeEEPROM(x+32,ports[x].address >> 3)) { printf("\nEEPROM write error:  write operation timed out.\n\n"); return; }
			if (!writeEEPROM(x+36,ports[x].intr)) { printf("\nEEPROM write error:  write operation timed out.\n\n"); return; }
		}
		else
		{
			if (!writeEEPROM(x,ports[x].address >> 3)) { printf("\nEEPROM write error:  write operation timed out.\n\n"); return; }
			if (!writeEEPROM(x+8,ports[x].intr)) { printf("\nEEPROM write error:  write operation timed out.\n\n"); return; }
		}

        printf(".");
    }

    printf("Done\n\n");
}

void readConfigFromEEPROM()
{
    int proto1, proto2;

    proto1 = 0;
    proto2 = 0;

    if (base_address == 0)
    {
        printf("Enter Base Address in HEX: ");
        scanf("%x",&base_address);
    }

    printf("Reading configuration from EEPROM at base address 0x%x",base_address);

	if ( board_type == 2 )
	{
        //=======NEW for EMM8P======//
        if (!readEEPROM(40, &proto1)) { printf("\nEEPROM read error:  read operation timed out.\n\n"); return; } ;
	}
	else
	{
        //=======NEW for EMM8P======//
        if (!readEEPROM(16, &proto1)) { printf("\nEEPROM read error:  read operation timed out.\n\n"); return; } ;
        if (!readEEPROM(17, &proto2)) { printf("\nEEPROM read error:  read operation timed out.\n\n"); return; } ;
	}

        //=======END NEW for EMM8P======//

    for (int x = 0; x < ports_variable; x++)
    {
		if ( board_type == 2 )
		{
	        if (!readEEPROM(x+32,&(ports[x].address))) { printf("\nEEPROM read error:  read operation timed out.\n\n"); return; } ;
		    if (!readEEPROM(x+36,&(ports[x].intr))) { printf("\nEEPROM read error:  read operation timed out.\n\n"); return; };
		}
		else
		{
	        if (!readEEPROM(x,&(ports[x].address))) { printf("\nEEPROM read error:  read operation timed out.\n\n"); return; } ;
		    if (!readEEPROM(x+8,&(ports[x].intr))) { printf("\nEEPROM read error:  read operation timed out.\n\n"); return; };
		}

        ports[x].address = ports[x].address << 3;

        if( x < 4 )
            ports[x].serialproto = (proto1 >> (x*2)) & 0x3;
        else
            ports[x].serialproto = (proto2 >> ((x-4)*2)) & 0x3;

        printf(".");
    }
    printf("Done\n\n");
    //if (!commandline) reconfigEMM8FromMemory();

    configured = 1;
}

void displayConf()
{
    if (!configured)
    {
        return;
    }

    printf("Base Address: 0x%x\n",base_address);
    printf("Port\tAddress\tInterrupt\tSerialProto\n");
    for (int x = 0; x < ports_variable; x++)
    {
            printf(" %d\t 0x%x\t    %d\t    %s\n",(int)x,(int)ports[x].address, (int)ports[x].intr, (ports[x].serialproto == SERIAL_RS232)?"RS232": ((ports[x].serialproto == SERIAL_RS422)?"RS422": "RS485") );
    }
}

void displayEEPROMConf()
{
    int proto1, proto2, tempProto;

    proto1 = 0;
    proto2 = 0;

    int addr = 0;
    int intr = 0;


    if (base_address == 0)
    {
        printf("Enter Base Address in HEX: ");
        scanf("%x",&base_address);
    }

    printf("\nEEPROM config at 0x%x:\n", (int)base_address);
    printf("Port\tAddress\tInterrupt  SerialProto\n");

	if ( board_type == 2 )
	{
	    if (!readEEPROM(40,&proto1)) { printf("\nEEPROM read error:  read operation timed out.\n\n"); return; } ;
	}
	else
	{
	    if (!readEEPROM(16,&proto1)) { printf("\nEEPROM read error:  read operation timed out.\n\n"); return; } ;
		if (!readEEPROM(17,&proto2)) { printf("\nEEPROM read error:  read operation timed out.\n\n"); return; } ;
	}


    for (int x = 0; x < ports_variable; x++)
    {
		if ( board_type == 2 )
		{
			if (!readEEPROM(x+32,&addr)) { printf("\nEEPROM read error:  read operation timed out.\n\n"); return; } ;
			if (!readEEPROM(x+36,&intr)) { printf("\nEEPROM read error:  read operation timed out.\n\n"); return; };
		}
		else
		{
			if (!readEEPROM(x,&addr)) { printf("\nEEPROM read error:  read operation timed out.\n\n"); return; } ;
			if (!readEEPROM(x+8,&intr)) { printf("\nEEPROM read error:  read operation timed out.\n\n"); return; };
		}


      addr = addr << 3;

      if( x < 4 )
          tempProto = (proto1 >> (x*2)) & 0x3;
      else
          tempProto = (proto2 >> ((x-4)*2)) & 0x3;

            printf(" %d\t 0x%x\t    %d\t    %s\n",(int)x,(int)addr,(int)intr, (tempProto == SERIAL_RS232)?"RS232": ((tempProto == SERIAL_RS422)?"RS422": "RS485") );
    }
}

void enterConf()
{

    printf("Enter Base Address in HEX: ");
    scanf("%x",&base_address);
    printf("\n");
//   printf("If configuring EMM8P enter 1, else 0: ");
//   scanf("%d", &conf);

    printf("\n");

    for (int x = 0; x < ports_variable; x++)
    {
        printf("Port %d:\n",x);
        printf("Enter Address in HEX: ");
        scanf("%x",&(ports[x].address));
        printf("Enter interrupt in DEC: ");
        scanf("%d",&(ports[x].intr));

		if ( board_type == 2 ) // if it is Neptune board
		{
			if( x == ports_variable - 1 )
			{
				printf("Enter serial protocol in DEC (%d=RS232, %d=RS422, %d=RS485): ", SERIAL_RS232, SERIAL_RS422, SERIAL_RS485);
				scanf("%d", &(ports[x].serialproto));
			}
		}
		else
		{
			if( board_type == 1 ) // if it is EMM8P board
			{
				printf("Enter serial protocol in DEC (%d=RS232, %d=RS422, %d=RS485): ", SERIAL_RS232, SERIAL_RS422, SERIAL_RS485);
				scanf("%d", &(ports[x].serialproto));
			}
		}
    }

    configured = 1;
}

int compare(char* str1, char* str2)
{
     int pos = 0;

    while (str1[pos] != 0)
    {
        if (str1[pos] != str2[pos]) return 0;
      pos++;
    }
    if (str2[pos] != 0) return 0;
    return 1;
}

void bAdd(char* a)
{
     if (a != NULL) { sscanf(a,"%x",&base_address); }
}

void instr()
{
    printf("Usage (base_addressr = optional argument, base address of board in HEX):\n");
    printf("eepconf -d base_addressr             Display EEPROM Contents\n");
    printf("eepconf -e base_addressr             Load configuration from EEPROM to Board\n");
    printf("eepconf -lf filename            Load configuration from File to Board\n");
    printf("eepconf -le filename base_addressr   Load configuration from File to Board+EEPROM\n");
    printf("eepconf -se filename base_addressr   Save configuration from EEPROM to File\n");
    exit(0);
}

void printmenu1()
{
      printf("\n1) Display current EEPROM configuration\n");
      printf("2) Load configuration to Memory manually\n");
      printf("3) Load configuration from File to Memory\n");
      printf("4) Load configuration from EEPROM to Memory\n");
      printf("5) Load configuration from Memory to Board\n");
      printf("6) Load configuration from EEPROM to Board\n");
      printf("7) Save configuration from Memory to File\n");
      printf("8) Save configuration from Memory to EEPROM\n");
      printf("9) Change board base address (currently 0x%x)\n",base_address);
        printf("Q) Quit\n");
      printf("\n> ");
}

void printmenu2()
{
      printf("\n1) Display EEPROM contents\n");
      printf("2) User Input ->   Memory\n");
      printf("3) File       ->   Memory\n");
      printf("4) EEPROM     ->   Memory\n");
      printf("5) Memory     ->   Board Registers\n");
      printf("6) EEPROM     ->   Board Registers\n");
      printf("7) Memory     ->   File\n");
      printf("8) Memory     ->   EEPROM\n");
      printf("9) Change board base address (currently 0x%x)\n",base_address);
        printf("Q) Quit\n");
      printf("\n> ");
}

int main(int argc, char ** argv)
{
	char buf[256];
	int init = 0;
	int conf;

    printf("\nEEPCONF.EXE   Emerald-MM series / Neptune SBC Configuration Program\n");
    printf("v2.1 3-02-2009      (c) 2009 Diamond Systems Corp.\n\n");

    #if defined(linux)
    iopl(3);
    printf("in here\n");
    #elif defined(_WIN32)
    DSCCB dsccb;
    DSCB dscb;

    dsccb.base_address = 0x100;
    dsccb.int_level = 5;

    int result =  dscInit(DSC_VERSION);
    result += dscInitBoard(DSC_RAW, &dsccb, &dscb);

    if( result != 0 )
        printf("DSCUD Error. Please check DSCUD version and installation\n");

    #endif

   printf("Enter board being used (0 = EMM8, 1 = EMM8P, 2 = Neptune): ");
   scanf("%d", &conf);

   if ( conf == 2 )
   {
	   ports_variable = 4;
	   board_type = 2;
   }
   else if ( conf == 1 )
   {
	   ports_variable = 8;
	   board_type = 1;
   }
   else
   {
	   ports_variable = 8;
	   board_type = 0;
   }

    if (argc > 1)
    {
        commandline = 1;
        if (compare(argv[1],"/?") || compare(argv[1],"-help"))
      {
            instr();
      }

      if (compare(argv[1],"-d"))
      {
          if (argc > 2) bAdd(argv[2]);
         printf("\n");
           readConfigFromEEPROM();
         displayConf();
         exit(0);
      }

      if (compare(argv[1],"-e"))
      {
          if (argc > 2) bAdd(argv[2]);
         printf("\n");
        readConfigFromEEPROM();
         displayConf();
         reconfigEMM8FromMemory();
         exit(0);
      }

      if (compare(argv[1],"-lf"))
      {
          if (argc > 3) bAdd(argv[3]);
         printf("\n");
           loadConfigFromFile(argv[2]);
         displayConf();
         reconfigEMM8FromMemory();
         exit(0);
      }

      if (compare(argv[1],"-le"))
      {
          if (argc > 3) bAdd(argv[3]);
         printf("\n");
           loadConfigFromFile(argv[2]);
         displayConf();
         writeConfigToEEPROM();
         reconfigEMM8FromMemory();
         exit(0);
      }

      if (compare(argv[1],"-se"))
      {
          if (argc > 3) bAdd(argv[3]);
         printf("\n");
           readConfigFromEEPROM();
         displayConf();
         saveConfigToFile(argv[2]);
         exit(0);
      }

      instr();
    }

    char filename[100] = {0};
    char key = 0;

    while (1)
    {
      if (init == 0)
      {
          init = 1;
         key = '9';
      }
      else
      {
         if (key != '1') displayConf();

        printmenu2();

        #if defined(linux) || defined(_WIN32)
              fgets(buf, 256, stdin);
            key = buf[0];
        #else
            key = getch();
        #endif

          printf("\n\n");
      }

      switch (key)
      {
      case '1':
          displayEEPROMConf();
         break;

      case '2':
          enterConf();
         break;

      case '3':
          printf("\nEnter filename: ");
            scanf("%s",filename);
         loadConfigFromFile(filename);
          break;

      case '4':
         readConfigFromEEPROM();
         break;

      case '5':
          reconfigEMM8FromMemory();
         break;

      case '6':
          readConfigFromEEPROM();
         reconfigEMM8FromMemory();
         break;

      case '7':
          printf("\nEnter filename: ");
            scanf("%s",filename);
         saveConfigToFile(filename);
          break;

      case '8':
          writeConfigToEEPROM();
         break;

      case '9':
          printf("Enter base address in HEX: ");
         scanf("%x",&base_address);
         break;

      case 'q':
          exit(0);
            break;
      }
    }
    return 0;
}

